import SimpleITK as sitk
import ants
import numpy as np
from scipy.ndimage import binary_fill_holes

from liom_toolkit.segmentation import remove_small_structures


def segment_3d_brain(volume: ants.ANTsImage, k: int = 5, use_log: bool = True,
                     threshold_method: str = "otsu") -> np.ndarray:
    """
    Segment a 3D brain volume using a watershed algorithm.
    Source: https://github.com/linum-uqam/sbh-reconstruction/blob/51271c84347afccb21483cfd3fcbde77d537929c/slicercode/segmentation/brainMask.py

    :param volume: The volume to segment.
    :type volume: ants.ANTsImage
    :param k: The size of the median filter.
    :type k: int
    :param use_log: Whether to use the log of the volume.
    :type use_log: bool
    :param threshold_method: The threshold method to use. Either "otsu" or "triangle".
    :type threshold_method: str
    :return: The segmented mask
    :rtype: np.ndarray
    """
    np_volume = volume.numpy()
    vol_p = np.copy(np_volume)
    if use_log:
        vol_p[np_volume > 0] = np.log(vol_p[np_volume > 0])

    # Creating a sitk image + smoothing
    img = sitk.GetImageFromArray(vol_p)
    img = sitk.Median(img, [k, k, k])

    # Segmenting using an Otsu threshold
    if threshold_method == "otsu":
        marker_img = ~sitk.OtsuThreshold(img)
    elif threshold_method == "triangle":
        marker_img = ~sitk.TriangleThreshold(img)
    else:
        marker_img = ~sitk.OtsuThreshold(img)

    # Using a watershed algorithm to optimize the mask
    ws = sitk.MorphologicalWatershedFromMarkers(img, marker_img)

    # Separating into foreground / background
    seg = sitk.ConnectedComponent(ws != ws[0, 0, 0])

    # Filling holes and returning the mask
    mask = fill_holes_2d_3d(sitk.GetArrayFromImage(seg))

    # Remove small objects
    mask = remove_small_structures(vol_p, mask)

    return mask


def fill_holes_2d_3d(mask: np.ndarray) -> np.ndarray:
    """
    Fill holes in a 2D and 3D mask.
    Source: https://github.com/linum-uqam/sbh-reconstruction/blob/51271c84347afccb21483cfd3fcbde77d537929c/slicercode/segmentation/brainMask.py

    :param mask: The mask to fill holes in.
    :type mask: np.ndarray
    :return: The mask with holes filled.
    :rtype: np.ndarray
    """
    # Filling holes and returning the mask
    mask = binary_fill_holes(mask)

    # Fill holes (in 2D)
    nx, ny, nz = mask.shape
    for x in range(nx):
        mask[x, :, :] = binary_fill_holes(mask[x, :, :])
    for y in range(ny):
        mask[:, y, :] = binary_fill_holes(mask[:, y, :])
    for z in range(nz):
        mask[:, :, z] = binary_fill_holes(mask[:, :, z])

    # Refill holes in 3D (in case some were missed)
    mask = binary_fill_holes(mask)
    return mask



import os

import ants
import ants.core.ants_image as iio
import numpy as np
from allensdk.core.reference_space_cache import ReferenceSpaceCache
from ants import utils


def convert_allen_nrrd_to_ants(volume: np.ndarray, resolution: float) -> ants.ANTsImage:
    """
    Convert a nrrd file form the Allen reference spaces to an ants image. The returned image will be in RAS+ orientation.

    :param volume: The already loaded nrrd file.
    :type volume: np.ndarray
    :param resolution: The resolution of the nrrd file in millimeters.
    :type resolution: float
    :return: The converted image.
    :rtype: ants.ANTsImage
    """
    # Set axis to RAS
    volume = np.moveaxis(volume, [0, 1, 2], [1, 2, 0])

    # Convert to ants image and set direction and spacing
    volume = ants.from_numpy(volume.astype("uint32"))
    volume.set_direction([[1., 0., 0.], [0., 1., 0.], [0., 0., -1.]])
    volume.set_spacing([resolution, resolution, resolution])

    return volume


def construct_reference_space_cache(resolution: int = 25) -> ReferenceSpaceCache:
    """
    Construct a reference space cache for the Allen brain atlas. Will use the 2017 adult version of the atlas.

    :param resolution: The resolution of the atlas in micron. Must be 10, 25, 50 or 100 microns
    :type resolution: int
    :return: The reference space cache.
    :rtype: ReferenceSpaceCache
    """
    # Check the resolution
    assert resolution in [10, 25, 50, 100], "Resolution must be 10, 25, 50 or 100"

    # Construct the reference space cache
    rsc = ReferenceSpaceCache(resolution=resolution, reference_space_key="annotation/ccf_2017")

    return rsc


def download_allen_template(data_dir: str, resolution: int = 25, keep_nrrd: bool = False,
                            rsc: ReferenceSpaceCache = None) -> ants.ANTsImage:
    """
    Download the allen mouse brain template in RAS+ orientation.

    :param data_dir: The directory to save the template to.
    :type data_dir: str
    :param resolution: The template resolution in micron. Must be 10, 25, 50 or 100 microns
    :type resolution: int
    :param keep_nrrd: Whether to keep the nrrd file or not.
    :type keep_nrrd: bool
    :param rsc: The reference space cache to use. If None, a new one will be constructed.
    :type rsc: ReferenceSpaceCache
    :return: The template as an ants image.
    :rtype: ants.ANTsImage
    """
    # Check the resolution
    assert int(resolution) in [10, 25, 50, 100], "Resolution must be 10, 25, 50 or 100"

    # filename
    nrrd_file = f"{data_dir}/allen_template_{resolution}.nrrd"

    # Downloading the template
    if rsc is None:
        rsc = construct_reference_space_cache(resolution=resolution)
    vol, metadata = rsc.get_template_volume(file_name=str(nrrd_file))

    ants_image = convert_allen_nrrd_to_ants(vol, resolution / 1000)

    # Remove nrrd file if unwanted
    if not keep_nrrd:
        os.remove(nrrd_file)

    return ants_image


def download_allen_atlas(data_dir: str, resolution: int = 25, keep_nrrd: bool = False,
                         rsc: ReferenceSpaceCache = None) -> ants.ANTsImage:
    """
    Download the allen mouse brain atlas and reorient it to RAS+.

    :param data_dir: The directory to save the atlas to.
    :type data_dir: str
    :param resolution: The resolution of the atlas in micron. Must be 10, 25, 50 or 100 microns
    :param resolution: int
    :param keep_nrrd: Whether to keep the nrrd file or not.
    :param keep_nrrd: bool
    :param rsc: The reference space cache to use. If None, a new one will be constructed.
    :param rsc: ReferenceSpaceCache
    :return: The atlas as an ants image.
    :rtype: ants.ANTsImage
    """
    assert resolution in [10, 25, 50, 100], "Resolution must be 10, 25, 50 or 100"
    # Download resolution is 10 micron to fix wrong region labels

    # Temporary filename
    nrrd_file = f"{data_dir}/allen_atlas_{resolution}.nrrd"

    # Downloading the annotation (if not already downloaded)
    if rsc is None:
        rsc = construct_reference_space_cache(resolution=resolution)
    vol, metadata = rsc.get_annotation_volume(str(nrrd_file))

    # Convert to ants image
    ants_image = convert_allen_nrrd_to_ants(vol, resolution / 1000)

    # Remove nrrd file if unwanted
    if not keep_nrrd:
        os.remove(nrrd_file)

    return ants_image


def apply_transforms(fixed, moving, transformlist,
                     interpolator='linear', imagetype=0,
                     whichtoinvert=None, compose=None,
                     defaultvalue=0, verbose=False, **kwargs):
    """
    FIXME: Function temporarily copied from ANTsPy. Will be removed once the issue with applying the transforms with double precision is resolved.

    Apply a transform list to map an image from one domain to another.
    In image registration, one computes mappings between (usually) pairs
    of images. These transforms are often a sequence of increasingly
    complex maps, e.g. from translation, to rigid, to affine to deformation.
    The list of such transforms is passed to this function to interpolate one
    image domain into the next image domain, as below. The order matters
    strongly and the user is advised to familiarize with the standards
    established in examples.

    ANTsR function: `antsApplyTransforms`

    Arguments
    ---------
    fixed : ANTsImage
        fixed image defining domain into which the moving image is transformed.

    moving : AntsImage
        moving image to be mapped to fixed space.

    transformlist : list of strings
        list of transforms generated by ants.registration where each transform is a filename.

    interpolator : string
        Choice of interpolator. Supports partial matching.
            linear
            nearestNeighbor
            multiLabel for label images (deprecated, prefer genericLabel)
            gaussian
            bSpline
            cosineWindowedSinc
            welchWindowedSinc
            hammingWindowedSinc
            lanczosWindowedSinc
            genericLabel use this for label images

    imagetype : integer
        choose 0/1/2/3 mapping to scalar/vector/tensor/time-series

    whichtoinvert : list of booleans (optional)
        Must be same length as transformlist.
        whichtoinvert[i] is True if transformlist[i] is a matrix,
        and the matrix should be inverted. If transformlist[i] is a
        warp field, whichtoinvert[i] must be False.
        If the transform list is a matrix followed by a warp field,
        whichtoinvert defaults to (True,False). Otherwise it defaults
        to [False]*len(transformlist)).

    compose : string (optional)
        if it is a string pointing to a valid file location,
        this will force the function to return a composite transformation filename.

    defaultvalue : scalar
        Default voxel value for mappings outside the image domain.

    verbose : boolean
        print command and run verbose application of transform.

    kwargs : keyword arguments
        extra parameters

    Returns
    -------
    ANTsImage or string (transformation filename)

    Example
    -------
    >>> import ants
    >>> fixed = ants.image_read( ants.get_ants_data('r16') )
    >>> moving = ants.image_read( ants.get_ants_data('r64') )
    >>> fixed = ants.resample_image(fixed, (64,64), 1, 0)
    >>> moving = ants.resample_image(moving, (64,64), 1, 0)
    >>> mytx = ants.registration(fixed=fixed , moving=moving ,
                                 type_of_transform = 'SyN' )
    >>> mywarpedimage = ants.apply_transforms( fixed=fixed, moving=moving,
                                               transformlist=mytx['fwdtransforms'] )
    """

    if not isinstance(transformlist, (tuple, list)) and (transformlist is not None):
        transformlist = [transformlist]

    accepted_interpolators = {"linear", "nearestNeighbor", "multiLabel", "gaussian",
                              "bSpline", "cosineWindowedSinc", "welchWindowedSinc",
                              "hammingWindowedSinc", "lanczosWindowedSinc", "genericLabel"}

    if interpolator not in accepted_interpolators:
        raise ValueError('interpolator not supported - see %s' % accepted_interpolators)

    args = [fixed, moving, transformlist, interpolator]

    if not isinstance(fixed, str):
        if isinstance(fixed, iio.ANTsImage) and isinstance(moving, iio.ANTsImage):
            for tl_path in transformlist:
                if not os.path.exists(tl_path):
                    raise Exception('Transform %s does not exist' % tl_path)

            inpixeltype = fixed.pixeltype
            fixed = fixed.clone('double')
            moving = moving.clone('double')
            warpedmovout = moving.clone('double')
            f = fixed
            m = moving
            if (moving.dimension == 4) and (fixed.dimension == 3) and (imagetype == 0):
                raise Exception('Set imagetype 3 to transform time series images.')

            wmo = warpedmovout
            mytx = []
            if whichtoinvert is None or (
                    isinstance(whichtoinvert, (tuple, list)) and (sum([w is not None for w in whichtoinvert]) == 0)):
                if (len(transformlist) == 2) and ('.mat' in transformlist[0]) and ('.mat' not in transformlist[1]):
                    whichtoinvert = (True, False)
                else:
                    whichtoinvert = tuple([False] * len(transformlist))

            if len(whichtoinvert) != len(transformlist):
                raise ValueError('Transform list and inversion list must be the same length')

            for i in range(len(transformlist)):
                ismat = False
                if '.mat' in transformlist[i]:
                    ismat = True
                if whichtoinvert[i] and (not ismat):
                    raise ValueError(
                        'Cannot invert transform %i (%s) because it is not a matrix' % (i, transformlist[i]))
                if whichtoinvert[i]:
                    mytx = mytx + ['-t', '[%s,1]' % (transformlist[i])]
                else:
                    mytx = mytx + ['-t', transformlist[i]]

            if compose is None:
                args = ['-d', fixed.dimension,
                        '-i', m,
                        '-o', wmo,
                        '-r', f,
                        '-n', interpolator]
                args = args + mytx
            if compose:
                tfn = '%scomptx.nii.gz' % compose if not compose.endswith('.h5') else compose
            else:
                tfn = 'NA'
            if compose is not None:
                mycompo = '[%s,1]' % tfn
                args = ['-d', fixed.dimension,
                        '-i', m,
                        '-o', mycompo,
                        '-r', f,
                        '-n', interpolator]
                args = args + mytx

            myargs = utils._int_antsProcessArguments(args)

            myverb = int(verbose)
            if verbose:
                print(myargs)

            processed_args = myargs + ['-z', str(1), '-v', str(myverb), str(1), '-e', str(imagetype), '-f',
                                       str(defaultvalue)]
            libfn = utils.get_lib_fn('antsApplyTransforms')
            libfn(processed_args)

            if compose is None:
                return warpedmovout.clone(inpixeltype)
            else:
                if os.path.exists(tfn):
                    return tfn
                else:
                    return None

        else:
            return 1
    else:
        args = args + ['-z', 1, '--float', 1, '-e', imagetype, '-f', defaultvalue]
        processed_args = utils._int_antsProcessArguments(args)
        libfn = utils.get_lib_fn('antsApplyTransforms')
        libfn(processed_args)



import os
import tempfile

import ants
import h5py as h5
import nibabel as nib
import nrrd
import numpy as np
import pywt
import zarr
from ome_zarr.io import parse_url
from ome_zarr.writer import write_image
from tqdm.auto import tqdm

from liom_toolkit.registration import align_annotations_to_volume
from liom_toolkit.utils import load_zarr, load_zarr_image_from_node, save_atlas_to_zarr, \
    CustomScaler, create_transformation_dict, generate_axes_dict, create_mask_from_zarr, save_label_to_zarr, \
    generate_label_color_dict_mask, load_node_by_name


def remove_stripe_based_wavelet_fft(image: np.ndarray, level: int = 5, sigma: int = 1, order: int = 8,
                                    pad: int = 150) -> np.ndarray:
    """
    Function gracefully taken from https://github.com/nghia-vo/sarepy/blob/master/sarepy/prep/stripe_removal_former.py
    Remove stripes using the method in https://doi.org/10.1364/OE.17.008567
    Angular direction is along the axis 0.

    Code adapted from tomopy source code https://github.com/tomopy/tomopy
    with a small improvement of using different ways of padding to
    reduce the side effect of the Fourier transform.


    :param image: The image to remove stripes from.
    :type image: np.ndarray
    :param level: Wavelet decomposition level.
    :type level: int
    :param sigma: Damping parameter. Larger is stronger.
    :type sigma: int
    :param order: Order of the the Daubechies wavelets.
    :type order: int
    :param pad: Padding for FFT
    :type pad: int
    :return: 2D array. Stripe-removed sinogram.
    :rtype: np.ndarray
    """
    (nrow, ncol) = image.shape
    if pad > 0:
        image = np.pad(image, ((pad, pad), (0, 0)), mode='mean')
        image = np.pad(image, ((0, 0), (pad, pad)), mode='edge')
    # Wavelet decomposition
    cH = []
    cV = []
    cD = []
    waveletname = "db" + str(order)
    for j in range(level):
        image, (cHt, cVt, cDt) = pywt.dwt2(image, waveletname)
        cH.append(cHt)
        cV.append(cVt)
        cD.append(cDt)
    for j in range(level):
        fcH = np.fft.fftshift(np.fft.fft2(cH[j]))  # cH for horizontal
        nrow1, ncol1 = fcH.shape
        y_hat = (np.arange(-nrow1, nrow1, 2, dtype=np.float32) + 1) / 2
        damp = 1 - np.exp(-np.power(y_hat, 2) / (2 * np.power(sigma, 2)))
        fcH = np.multiply(fcH, np.transpose(np.tile(damp, (ncol1, 1))))
        cH[j] = np.real(np.fft.ifft2(np.fft.ifftshift(fcH)))
    # Wavelet reconstruction.
    for j in range(level)[::-1]:
        image = image[0:cH[j].shape[0], 0:cH[j].shape[1]]
        image = pywt.idwt2((image, (cH[j], cV[j], cD[j])), waveletname)
    if pad > 0:
        image = image[pad:-pad, pad:-pad]
    return image[:nrow, :ncol]


def load_hdf5(hdf5_file: str, use_mem_map: bool = True, map_file: str = "temp.dat",
              base_key: str = "reconstructed_frame") -> np.ndarray:
    """
    Load the data from a HDF5 file. If use_mem_map is True, the data will be saved to a memmap file to save memory.

    :param hdf5_file: The HDF5 file to load.
    :type hdf5_file: str
    :param use_mem_map: Whether to use a memmap or not.
    :type use_mem_map: bool
    :param map_file: The memmap file to save to.
    :type map_file: str
    :param base_key: The base key of the HDF5 key list.
    :type base_key: str
    :return: The data from the HDF5 file.
    :rtype: np.ndarray
    """
    with h5.File(hdf5_file) as file:
        keys = file.keys()
        n_frames = len(keys)
        key_list = [base_key + "{:03d}".format(i + 1) for i in range(len(keys))]
        frame = file[key_list[0]][:]
        if use_mem_map:
            data = np.memmap(map_file, dtype=np.uint32, mode='w+', shape=(n_frames, frame.shape[0], frame.shape[1]))
        else:
            data = np.zeros((n_frames, frame.shape[0], frame.shape[1]), dtype=np.uint32)
        for i, key in enumerate(
                tqdm(key_list, desc="Loading HDF5 file..", unit=" frames", total=len(key_list),
                     leave=False, position=1)):
            frame = file[key][:]
            data[i, :, :] = frame

    return data


def convert_hdf5_to_nifti(hdf5_file: str, nifti_file: str, use_mem_map: bool = True,
                          base_key: str = "reconstructed_frame") -> None:
    """
    Convert a HDF5 file to a NIFTI file.

    :param hdf5_file: Path to the HDF5 file.
    :type hdf5_file: str
    :param nifti_file: Path to the NIFTI file.
    :type nifti_file: str
    :param use_mem_map: Whether to use a memmap or not.
    :type use_mem_map: bool
    :param base_key: The base key of the HDF5 key list.
    :type base_key: str
    """
    map_file = "temp.dat"
    data = load_hdf5(hdf5_file, use_mem_map, map_file, base_key=base_key)

    print("Saving...")
    ni_img = nib.Nifti1Image(data, affine=np.eye(4), dtype=np.uint16)
    nib.save(ni_img, nifti_file)
    print("Done!")
    if use_mem_map:
        os.remove(map_file)


def save_zarr(data: np.ndarray, zarr_file: str, remove_stripes: bool = False, scales: tuple = (6.5, 6.5, 6.5),
              chunks: tuple = (128, 128, 128)) -> None:
    """
    Save a numpy array to a zarr file.

    :param data: The data to save.
    :type data: np.ndarray
    :param zarr_file: The zarr file to save to.
    :type zarr_file: str
    :param remove_stripes: Whether to remove stripes from the data.
    :type remove_stripes: bool
    :param scales: The resolution of the image, in z y x order.
    :type scales: tuple
    :param chunks: The chunk size to use.
    :type chunks: tuple
    """
    n_dims = len(data.shape)
    if remove_stripes:
        for i in tqdm(range(data.shape[0]), desc="Removing stripes", leave=False, unit="frames",
                      total=data.shape[0], position=1):
            data[i, :, :] = remove_stripe_based_wavelet_fft(data[i, :, :])

    print("Saving...")
    os.mkdir(zarr_file)
    store = parse_url(zarr_file, mode="w").store
    root = zarr.group(store=store)
    write_image(image=data, group=root, axes=generate_axes_dict(n_dims),
                coordinate_transformations=create_transformation_dict(scales, 5, n_dims),
                storage_options=dict(chunks=chunks),
                scaler=CustomScaler(order=1, anti_aliasing=True, downscale=2, method="nearest", input_layer=0))
    print("Done!")


def convert_hdf5_to_zarr(hdf5_file: str, zarr_file: str, use_memmap: bool = True, remove_stripes: bool = False,
                         scales: tuple = (6.5, 6.5, 6.5), chunks: tuple = (128, 128, 128),
                         base_key: str = "reconstructed_frame") -> None:
    """
    Convert a HDF5 file from the lightsheet microscope to a zarr file.

    :param hdf5_file: Path to the HDF5 file.
    :type hdf5_file: str
    :param zarr_file: Path to the zarr file.
    :type zarr_file: str
    :param use_memmap: Whether to use a memmap or not.
    :type use_memmap: bool
    :param remove_stripes: Whether to remove stripes from the data.
    :type remove_stripes: bool
    :param scales: The resolution of the image, in z y x order.
    :type scales: tuple
    :param chunks: The chunk size to use.
    :type chunks: tuple
    :param base_key: The base key of the HDF5 key list.
    :type base_key: str
    """

    map_file = "temp.dat"
    data = load_hdf5(hdf5_file, use_memmap, map_file, base_key=base_key)

    save_zarr(data, zarr_file, remove_stripes, scales=scales, chunks=chunks)
    if use_memmap:
        os.remove(map_file)


def convert_nifti_to_zarr(nifti_file: str, zarr_file: str, scales: tuple = (6.5, 6.5, 6.5),
                          chucks: tuple = (128, 128, 128), transpose: bool = False) -> None:
    """
    Convert a NIFTI file to a zarr file.

    :param nifti_file: The NIFTI file to convert.
    :type nifti_file: str
    :param zarr_file: The zarr file to save to.
    :type zarr_file: str
    :param scales: The resolution of the image, in z y x order.
    :type scales: tuple
    :param chucks: The chunk size to use in the zarr file.
    :type chucks: tuple
    :param transpose: Whether to transpose the data or not.
    :type transpose: bool
    """
    print("Loading...")
    ni_img = nib.load(nifti_file)
    data = ni_img.get_fdata()
    if transpose:
        data = np.transpose(data, (2, 1, 0))
    save_zarr(data, zarr_file, scales=scales, chunks=chucks)


def convert_nrrd_to_zarr(nrrd_file: str, zarr_file: str, scales: tuple = (6.5, 6.5, 6.5),
                         chucks: tuple = (128, 128, 128)) -> None:
    """
    Convert a NRRD file to a zarr file.

    :param nrrd_file: The NRRD file to convert.
    :type nrrd_file: str
    :param zarr_file: The zarr file to save.
    :type zarr_file: str
    :param scales: The resolution of the image, in z y x order.
    :type scales: tuple
    :param chucks: The chunk size to use in the zarr file.
    :type chucks: tuple
    """
    print("Loading...")
    data, header = nrrd.read(nrrd_file)
    save_zarr(data, zarr_file, scales=scales, chunks=chucks)


def create_multichannel_zarr(auto_fluo_file: str, vascular_file: str, zarr_file: str, temp_dir: str | None,
                             scales: tuple = (6.5, 6.5, 6.5), chunks: tuple = (128, 128, 128),
                             use_mem_map: bool = False, remove_stripes: bool = False) -> None:
    """
    Create a multichannel zarr file from the auto-fluorescence and vascular data.

    :param auto_fluo_file: The path to the auto-fluorescence hdf5 file.
    :type auto_fluo_file: str
    :param vascular_file: The path to the vascular hdf5 file.
    :type vascular_file: str
    :param zarr_file: The path to the zarr file to save the volume to.
    :type zarr_file: str
    :param temp_dir: The temporary directory to save the memmap files to.
    :type temp_dir: str
    :param scales: The physical resolution of the volume per axis.
    :type scales: tuple
    :param chunks: The chunk size to use for the volume.
    :type chunks: tuple
    :param use_mem_map: Whether to use a memory map for the hdf5 files.
    :type use_mem_map: bool
    :param remove_stripes: Whether to remove stripes from the volume.
    :type remove_stripes: bool
    :return:
    """
    # Extract data from the hdf5 files
    auto_fluo = load_hdf5(auto_fluo_file, use_mem_map=use_mem_map, map_file=f"{temp_dir}/647nm.dat")
    vascular = load_hdf5(vascular_file, use_mem_map=use_mem_map, map_file=f"{temp_dir}/555nm.dat")

    # Merge the data along a new fourth dimension at index 0
    volume = np.stack((auto_fluo, vascular), axis=0)

    # Save the volume to a zarr file
    save_zarr(volume, zarr_file, scales=scales, chunks=chunks, remove_stripes=remove_stripes)
    del auto_fluo, vascular, volume


def create_full_zarr_volume(auto_fluo_file: str, vascular_file: str, zarr_file: str, template_path: str,
                            scales: tuple = (6.5, 6.5, 6.5), chunks: tuple = (128, 128, 128), use_mem_map: bool = False,
                            remove_stripes: bool = False, original_volume_orientation: str = "RSP") -> None:
    """
    Create a full zarr volume from the auto-fluorescence and vascular data. The annotations will be aligned to the
    auto-fluorescence data and saved to the zarr file. The mask will also be created and saved to the zarr file.

    :param auto_fluo_file: The path to the auto-fluorescence hdf5 file.
    :type auto_fluo_file: str
    :param vascular_file: The path to the vascular hdf5 file.
    :type vascular_file: str
    :param zarr_file: The path to the zarr file to save the volume to.
    :type zarr_file: str
    :param template_path: The path to the template to align the annotations to.
    :type template_path: str
    :param scales: The physical resolution of the volume per axis.
    :type scales: tuple
    :param chunks: The chunk size to use for the volume.
    :type chunks: tuple
    :param use_mem_map: Whether to use a memory map for the hdf5 files.
    :type use_mem_map: bool
    :param remove_stripes: Whether to remove stripes from the volume.
    :type remove_stripes: bool
    :param original_volume_orientation: The original orientation of the volume.
    :type original_volume_orientation: str
    """
    temp_dir = tempfile.TemporaryDirectory()
    resolution_level = 2

    pbar = tqdm(total=5, desc="Creating zarr volume")
    pbar.set_postfix({"step": "Creating multichannel zarr"})
    create_multichannel_zarr(auto_fluo_file, vascular_file, zarr_file, temp_dir.name, scales=scales, chunks=chunks,
                             use_mem_map=use_mem_map, remove_stripes=remove_stripes)
    pbar.update(1)

    pbar.set_postfix({"step": "Creating temporary mask"})
    # Load image for image information
    nodes = load_zarr(zarr_file)
    target_image = load_zarr_image_from_node(nodes[0], resolution_level, channel=0)
    # Create the temporary mask
    mask = create_mask_from_zarr(zarr_file, resolution_level)
    mask = mask.astype("uint32")
    mask = ants.from_numpy(mask)
    mask.set_direction(target_image.direction)
    mask.set_spacing(target_image.spacing)
    mask.set_origin(target_image.origin)

    pbar.update(1)

    pbar.set_postfix({"step": "Aligning annotations to volume"})
    # Align the annotations to the volume
    nodes = load_zarr(zarr_file)
    target_image = load_zarr_image_from_node(nodes[0], resolution_level, channel=0)
    template = ants.image_read(template_path)

    atlas = align_annotations_to_volume(target_volume=target_image, mask=mask, template=template, resolution=25,
                                        keep_intermediary=False, data_dir=temp_dir.name)

    save_atlas_to_zarr(zarr_file, atlas, scales=scales, chunks=chunks, resolution_level=resolution_level,
                       orientation=original_volume_orientation)
    temp_dir.cleanup()
    pbar.update(1)

    # Creating final mask
    pbar.set_postfix({"step": "Creating final mask"})
    nodes = load_zarr(zarr_file)
    atlas_node = load_node_by_name(nodes, "atlas")
    atlas = load_zarr_image_from_node(atlas_node, resolution_level)

    # Set all non-zero pixels of the atlas to 1
    new_mask = atlas.numpy()
    new_mask[new_mask > 0] = 1

    # Save to zarr
    new_mask = np.transpose(new_mask, (2, 1, 0))
    new_mask = new_mask.astype("int8")
    color_dict = generate_label_color_dict_mask()
    save_label_to_zarr(new_mask, zarr_file, scales=scales, chunks=chunks, color_dict=color_dict,
                       name="mask", resolution_level=resolution_level)
    pbar.update(1)

    pbar.set_postfix({"step": "Done"})
    pbar.update(1)
    pbar.close()
